/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-    */
/* ex: set filetype=cpp softtabstop=4 shiftwidth=4 tabstop=4 cindent expandtab: */

/*
  $Id: SimpleBehavior.cpp 2448 2011-04-09 15:31:41Z adeguet1 $

  Author(s):	Balazs Vagvolgyi, Simon DiMaio, Anton Deguet
  Created on:	2008-05-23

  (C) Copyright 2008 Johns Hopkins University (JHU), All Rights
  Reserved.

--- begin cisst license - do not edit ---

This software is provided "as is" under an open source license, with
no warranty.  The complete license can be found in license.txt and
http://www.cisst.org/cisst/license.txt.

--- end cisst license ---
*/

#include <cisstOSAbstraction/osaThreadedLogFile.h>
#include <cisstOSAbstraction/osaSleep.h>
#include <cisstMultiTask/mtsTaskManager.h>
#include <cisstMultiTask/mtsInterfaceRequired.h>

#include <VoiceControl.h>

#define CONTEXT_DISTANCE "distance"
#define CONTEXT_MAP "map"
#define CONTEXT_MODEL "model"

#define QUIT_MODE "quit"

#define DISTANCE_START "start"
#define DISTANCE_RESET "reset"
#define DISTANCE_FREEZE "freeze"

#define MAP_ADD "add marker"
#define MAP_SHOW "show all"
#define MAP_HIDE "hide all"

#define MODEL_TOGGLE_SURFACE "surface"
#define MODEL_TOGGLE_TUMOR "tumor"
#define MODEL_TOGGLE_TREE "tree"


VoiceControl::VoiceControl(const std::string & taskName):
mtsComponent(taskName)
{
    // to communicate with the interface of the resource
    mtsInterfaceRequired * required = AddInterfaceRequired("Voice");
    if (required) {
        required->AddEventHandlerWrite(&VoiceControl::EventContextChangedHandler, this, "ContextChanged");
        required->AddEventHandlerWrite(&VoiceControl::EventWordRecognizedHandler, this, "WordRecognized");
    }

    mtsInterfaceProvided * provided = AddInterfaceProvided("MeasureTrigger");
    if (provided) {
        provided->AddEventVoid(measureStart, "Start");
        provided->AddEventVoid(measureStop, "Stop");
        provided->AddEventVoid(measureExit, "Exit");
    }

    provided = AddInterfaceProvided("MapTrigger");
    if (provided) {
        provided->AddEventVoid(mapDisplay, "Display");
        provided->AddEventVoid(mapAddMarker, "Add");
        provided->AddEventVoid(mapShow, "Show");
        provided->AddEventVoid(mapHide, "Hide");
        provided->AddEventVoid(mapExit, "Exit");
    }
/*
    mtsInterfaceProvided * provided2 = AddInterfaceProvided("DropMarker");
    if(provided2){
            provided2->AddEventVoid(dropMarker, "DropMark");
*/

}


void VoiceControl::Configure(const std::string & CMN_UNUSED(filename))
{
	speechToCommands =
        new cscSpeechToCommands("mySpeechToCommands");

	speechToCommands->SetMicrophoneNumber(3);
    cscContext * areYouTalkingToMe = speechToCommands->AddContext("areYouTalkingToMe");
	// areYouTalkingToMe->SetFiltering(true);
    cscContext * confirmStart = speechToCommands->AddContext("confirmStart");
    cscContext * selectMode = speechToCommands->AddContext("selectMode");
    cscContext * measurement = speechToCommands->AddContext(CONTEXT_DISTANCE);
    cscContext * map = speechToCommands->AddContext(CONTEXT_MAP);
    cscContext * model = speechToCommands->AddContext(CONTEXT_MODEL);
    
    areYouTalkingToMe->AddWordWithTransition("voice control", "confirmStart");
	areYouTalkingToMe->AddWordWithTransition("listen", "confirmStart");

	confirmStart->AddWordWithTransition("yes", "selectMode");
    confirmStart->AddWordWithTransition("no", "areYouTalkingToMe");

    selectMode->AddWordWithTransition(CONTEXT_MAP, CONTEXT_MAP);
    selectMode->AddWordWithTransition(CONTEXT_DISTANCE, CONTEXT_DISTANCE);
    //selectMode->AddWordWithTransition(CONTEXT_MODEL, CONTEXT_MODEL);
    selectMode->AddWordWithTransition("stop listening", "areYouTalkingToMe");

    measurement->AddWord(DISTANCE_START);
    measurement->AddWord(DISTANCE_RESET);
    measurement->AddWord(DISTANCE_FREEZE);
    measurement->AddWordWithTransition(QUIT_MODE, "selectMode");

    map->AddWord(MAP_ADD);
    map->AddWord(MAP_SHOW);
    map->AddWord(MAP_HIDE);
    map->AddWordWithTransition(QUIT_MODE, "selectMode");

    //model->AddWord(MODEL_TOGGLE_SURFACE);
    //model->AddWord(MODEL_TOGGLE_

    speechToCommands->SetCurrentContext("areYouTalkingToMe");
    speechToCommands->Configure(); // creates all configurations files for Sphinx

    mtsComponentManager * componentManager = mtsComponentManager::GetInstance();
    componentManager->AddComponent(speechToCommands);
	componentManager->Connect(this->GetName(), "Voice", "mySpeechToCommands", "Default");       
}


void VoiceControl::Run(void)
{
	std::cerr << "running" << std::endl;
	this->ProcessQueuedCommands();
	this->ProcessQueuedEvents();
}


void VoiceControl::EventContextChangedHandler(const mtsStdString & value)
{
	std::cout << "----------------" << value.Data << std::endl;

    std::string nextContext = value.Data;

    if (nextContext == CONTEXT_MAP)
        mapDisplay();
}

void VoiceControl::EventWordRecognizedHandler(const mtsStdString & value)
{
	std::cout << "++++++++++++++++" << value.Data << std::endl;

    std::string currContext = speechToCommands->GetCurrentContext()->GetName();
    std::string word = value.Data;

    if (currContext == CONTEXT_DISTANCE) {
        if ((word == DISTANCE_START) || (word == DISTANCE_RESET)) {
            measureStart();
        }
        else if (word == DISTANCE_FREEZE) {
            measureStop();
        }
        else if (word == QUIT_MODE) {
            measureExit();
        }
    }
    else if (currContext == CONTEXT_MAP) {
        if (word == MAP_ADD) {
            mapAddMarker();
        }
        else if (word == MAP_SHOW) {
            mapShow();
        }
        else if (word == MAP_HIDE) {
            mapHide();
        }
        else if (word == QUIT_MODE) {
            mapExit();
        }
    }
};
